/*********************************************************************
*
*  Copyright (c) 2009, Willow Garage, Inc.
*  All rights reserved.
*
*  Redistribution and use in source and binary forms, with or without
*  modification, are permitted provided that the following conditions
*  are met:
*
*   * Redistributions of source code must retain the above copyright
*     notice, this list of conditions and the following disclaimer.
*   * Redistributions in binary form must reproduce the above
*     copyright notice, this list of conditions and the following
*     disclaimer in the documentation and/or other materials provided
*     with the distribution.
*   * Neither the name of the Willow Garage nor the names of its
*     contributors may be used to endorse or promote products derived
*     from this software without specific prior written permission.
*
*  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
*  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
*  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
*  FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
*  COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
*  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
*  BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
*  LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
*  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
*  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
*  ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
*  POSSIBILITY OF SUCH DAMAGE.
*********************************************************************/

// Author(s): Matei Ciocarlie

#ifndef _GRASP_EXECUTOR_WITH_APPROACH_
#define _GRASP_EXECUTOR_WITH_APPROACH_

#include "object_manipulator/grasp_execution/grasp_executor.h"

namespace object_manipulator {

//! Uses an interpolated IK approach from pregrasp to final grasp
/*! Initial check consists of IK for the pre-grasp, followed by generation
  of an interpolated IK path from pre-grasp to grasp. This is then followed
  by computation of an interpolated IK path from grasp to lift.

  Execution consists of using move arm to go to the pre-grasp, then execution of 
  the interpolated IK paths for grasp. Lift is executed separately, called from
  the higher level executor.

  Note that we do not use the pre-grasp from the database directly; rather, the
  pre-grasp is obtained by backing up the grasp by a pre-defined amount. This
  allows us more flexibility in choosing the pre-grasp. However, we can no longer
  always assume the pre-grasp is collision free, which we could if we used the 
  pre-grasp from the database directly.

  In the most recent database version, the pre-grasp was obtained by backing up 
  10 cm, so we know at least that that is not colliding with the object. 
*/
class GraspExecutorWithApproach : public GraspExecutor
{
 protected:
  //! The result of interpolated IK from pre-grasp to grasp
  trajectory_msgs::JointTrajectory interpolated_grasp_trajectory_;

  //! Computes an interpolated IK trajectory between pre_grasp and grasp and checks if it has enough points
  bool getInterpolatedIKForGrasp(const object_manipulation_msgs::PickupGoal &pickup_goal,
				 const object_manipulation_msgs::Grasp &grasp,
				 trajectory_msgs::JointTrajectory &grasp_trajectory);

  //! Collision operations to be used when planning the grasp motion
  virtual motion_planning_msgs::OrderedCollisionOperations 
    collisionOperationsForGrasp(const object_manipulation_msgs::PickupGoal &pickup_goal);

  //! Dynamic link padding to be used for grasp operation
  virtual std::vector<motion_planning_msgs::LinkPadding> 
    linkPaddingForGrasp(const object_manipulation_msgs::PickupGoal &pickup_goal);

  //! Computes IK for the pre-grasp and interpolated IK from pre-grasp to grasp and from grasp to lift
  virtual bool prepareGrasp(const object_manipulation_msgs::PickupGoal &pickup_goal,
			    const object_manipulation_msgs::Grasp &grasp);
  
  //! Uses move_arm to get to pre-grasp, then executes the interpolated IK grasp path and lift path
  virtual Result executeGrasp(const object_manipulation_msgs::PickupGoal &pickup_goal,
			      const object_manipulation_msgs::Grasp &grasp);

 public:
  //! Alse adds a grasp marker at the pre-grasp location
 GraspExecutorWithApproach(GraspMarkerPublisher *pub) : GraspExecutor(pub) {}

  //! Retreats along the gripper approach direction
  virtual Result retreat(const object_manipulation_msgs::PickupGoal &pickup_goal);
};

} //namespace grasp_execution

#endif
