/*********************************************************************
*
*  Copyright (c) 2009, Willow Garage, Inc.
*  All rights reserved.
*
*  Redistribution and use in source and binary forms, with or without
*  modification, are permitted provided that the following conditions
*  are met:
*
*   * Redistributions of source code must retain the above copyright
*     notice, this list of conditions and the following disclaimer.
*   * Redistributions in binary form must reproduce the above
*     copyright notice, this list of conditions and the following
*     disclaimer in the documentation and/or other materials provided
*     with the distribution.
*   * Neither the name of the Willow Garage nor the names of its
*     contributors may be used to endorse or promote products derived
*     from this software without specific prior written permission.
*
*  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
*  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
*  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
*  FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
*  COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
*  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
*  BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
*  LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
*  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
*  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
*  ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
*  POSSIBILITY OF SUCH DAMAGE.
*********************************************************************/

// Author(s): Matei Ciocarlie

#include "object_manipulator/grasp_execution/grasp_executor.h"

#include "object_manipulator/tools/exceptions.h"
#include "object_manipulator/tools/hand_description.h"

namespace object_manipulator {

/*! If the object is a database model, its pose is returned.
  If the object is a cluster, the target pose if the origin, in the cloud's frame.
*/
geometry_msgs::PoseStamped GraspExecutor::getTargetPose(const object_manipulation_msgs::PickupGoal &pickup_goal)
{
  switch(pickup_goal.target.type)
    {
    case object_manipulation_msgs::GraspableObject::DATABASE_MODEL:
      return pickup_goal.target.model_pose.pose;
    case object_manipulation_msgs::GraspableObject::POINT_CLUSTER:
      {
	geometry_msgs::PoseStamped pose;
	pose.header = pickup_goal.target.cluster.header;
	pose.pose.position.x = pose.pose.position.y = pose.pose.position.z = 0.0;
	pose.pose.orientation.x = pose.pose.orientation.y = pose.pose.orientation.z = 0.0;
	pose.pose.orientation.w = 1.0;
	return pose;
      }
    default:
      ROS_ERROR("Unknown type of graspable object in getTargetPose");
      throw GraspException("Unknown type of graspable object in getTargetPose");
    }
}

/*! For debug, this is also used to modify the transform in some grasp-sepcific way, such as 
  moving it "back" (along negative X) some more.
*/
geometry_msgs::PoseStamped GraspExecutor::transformToModelFrame(const geometry_msgs::PoseStamped &target_pose, 
								const geometry_msgs::Pose &in_pose)
{
  tf::Transform model_trans;
  tf::poseMsgToTF(target_pose.pose, model_trans);
  tf::Transform in_trans;
  tf::poseMsgToTF(in_pose, in_trans);
  in_trans = model_trans * in_trans;

  //debug: move back some more  
  tf::Transform move_back_trans;
  move_back_trans.setIdentity();
  move_back_trans.setOrigin( btVector3(-0.001,0,0) );
  in_trans = in_trans * move_back_trans;
  
  geometry_msgs::PoseStamped out_pose;
  tf::poseTFToMsg(in_trans, out_pose.pose);
  out_pose.header = target_pose.header;
  return out_pose;
}

/*! Disables collision between gripper and target */
motion_planning_msgs::OrderedCollisionOperations 
GraspExecutor::collisionOperationsForLift(const object_manipulation_msgs::PickupGoal &pickup_goal)
{
  motion_planning_msgs::OrderedCollisionOperations ord;
  motion_planning_msgs::CollisionOperation coll;
  coll.object1 = handDescription().gripperCollisionName(pickup_goal.arm_name);
  coll.object2 = pickup_goal.collision_object_name;
  coll.operation = motion_planning_msgs::CollisionOperation::DISABLE;
  ord.collision_operations.push_back(coll);
  return ord;
}

std::vector<motion_planning_msgs::LinkPadding> 
GraspExecutor::fingertipPadding(const object_manipulation_msgs::PickupGoal &pickup_goal, double pad)
{
  std::vector<motion_planning_msgs::LinkPadding> padding_vec;
  motion_planning_msgs::LinkPadding padding;
  padding.padding = pad;
  
  std::vector<std::string> links = handDescription().fingertipLinks(pickup_goal.arm_name);
  for (size_t i=0; i<links.size(); i++)
  {
    padding.link_name = links[i];
    padding_vec.push_back(padding); 
  }  
  return padding_vec; 
}

/*! Zero padding on fingertip links */
std::vector<motion_planning_msgs::LinkPadding> 
GraspExecutor::linkPaddingForLift(const object_manipulation_msgs::PickupGoal &pickup_goal)
{
  return fingertipPadding(pickup_goal, 0.0);
}

bool GraspExecutor::getInterpolatedIKForLift(const object_manipulation_msgs::PickupGoal &pickup_goal,
					     const object_manipulation_msgs::Grasp &grasp,
					     const std::vector<double> &grasp_joint_angles,
					     trajectory_msgs::JointTrajectory &lift_trajectory)
{
  geometry_msgs::PoseStamped target_pose = getTargetPose(pickup_goal);
  geometry_msgs::PoseStamped grasp_pose = transformToModelFrame(target_pose, grasp.grasp_pose);
  grasp_pose.header.stamp = ros::Time(0);

  float actual_lift_distance;
  mechInterface().getInterpolatedIK(pickup_goal.arm_name, 
				    grasp_pose, 
				    pickup_goal.lift.direction, pickup_goal.lift.desired_distance,
				    grasp_joint_angles, 
				    grasp.grasp_posture, 
				    collisionOperationsForLift(pickup_goal), linkPaddingForLift(pickup_goal),
				    false, lift_trajectory, actual_lift_distance);
  ROS_INFO("  Lift distance: actual %f, min %f and desired %f", actual_lift_distance, pickup_goal.lift.min_distance,
           pickup_goal.lift.desired_distance);
  if (actual_lift_distance < pickup_goal.lift.min_distance)
  {
    ROS_INFO("  Lift trajectory  below min. threshold");
    if (marker_publisher_) marker_publisher_->colorGraspMarker(marker_id_, 0.0, 0.0, 1.0); //blue
    return false;
  }
  if (actual_lift_distance < pickup_goal.lift.desired_distance)
  {
    ROS_INFO("  Lift trajectory shorter than desired but above min. threshold");
  }
  else
  {
    ROS_INFO("  Lift trajectory matches desired value ");  
  }

  //check if the last pose in trajectory is acceptable
  //when we check for lift we use custom link padding, so we need to check here if the last pose 
  //is feasible with default padding; otherwise, move_arm might refuse to take us out of there
  if ( !mechInterface().checkStateValidity(pickup_goal.arm_name, lift_trajectory.points.back().positions) )
  {
    ROS_INFO("  Grasp executor: last pose in lift trajectory is unfeasible with default padding");
    return false;
  }

  return true;
}

GraspExecutor::Result GraspExecutor::lift(const object_manipulation_msgs::PickupGoal &pickup_goal)
{
  if (interpolated_lift_trajectory_.points.empty())
  {
    ROS_ERROR("  Grasp executor: lift trajectory not set");
    return UNFEASIBLE;
  }

  //execute the unnormalized interpolated trajectory from grasp to lift
  mechInterface().attemptTrajectory(pickup_goal.arm_name, interpolated_lift_trajectory_, true);
  return SUCCESS;
}

GraspExecutor::Result 
GraspExecutor::checkAndExecuteGrasp(const object_manipulation_msgs::PickupGoal &pickup_goal,
				    const object_manipulation_msgs::Grasp &grasp)
{
  //test if grasp is feasible and prepare trajectories  
  if (marker_publisher_)
  {
    geometry_msgs::PoseStamped target_pose = getTargetPose(pickup_goal);
    geometry_msgs::PoseStamped marker_pose = transformToModelFrame(target_pose, grasp.grasp_pose);
    marker_id_ = marker_publisher_->addGraspMarker(marker_pose);
  }  
  if (!prepareGrasp(pickup_goal, grasp)) return UNFEASIBLE;

  //execute grasp and check result
  Result grasp_result = executeGrasp(pickup_goal, grasp);
  if (grasp_result == UNFEASIBLE)
  {
    return UNFEASIBLE;
  } 
  else if (grasp_result == FAILED)
  {
    retreat(pickup_goal);
    return FAILED;
  }
  else if (grasp_result != SUCCESS)
  {
    throw GraspException("Unknown grasp execution result");
  }

  //check if there is anything in gripper; if not, open gripper and retreat

  if (!mechInterface().graspPostureQuery(pickup_goal.arm_name))
  {
    ROS_INFO("Hand reports that grasp was not successfully executed; releasing object and retreating");
    mechInterface().handPostureGraspAction(pickup_goal.arm_name, grasp,
					   object_manipulation_msgs::GraspHandPostureExecutionGoal::RELEASE);    
    retreat(pickup_goal);
    return FAILED;
  }
  
  //attach object to gripper
  mechInterface().attachObjectToGripper(pickup_goal.arm_name, pickup_goal.collision_object_name);

  //lift the object
  Result lift_result = lift(pickup_goal);
  if (lift_result == UNFEASIBLE)
  {
    //could not move gripper at all
    return LIFT_FAILED;
    //alternatively, we could still abort the grasp and retreat as the object has not been moved
    //retreat(pickup_goal);
    //return FAILED;
  }
  else if (lift_result == FAILED)
  {
    //we have lifted the object some, but not all the way up
    //we can not abort the grasp anymore, as we would drop the object
    return LIFT_FAILED;
  }
  else if (lift_result != SUCCESS)
  {
    throw GraspException("Unknown lift execution result");
  }

  return SUCCESS;
}

} //namespace object_manipulator
