/*********************************************************************
*
*  Copyright (c) 2009, Willow Garage, Inc.
*  All rights reserved.
*
*  Redistribution and use in source and binary forms, with or without
*  modification, are permitted provided that the following conditions
*  are met:
*
*   * Redistributions of source code must retain the above copyright
*     notice, this list of conditions and the following disclaimer.
*   * Redistributions in binary form must reproduce the above
*     copyright notice, this list of conditions and the following
*     disclaimer in the documentation and/or other materials provided
*     with the distribution.
*   * Neither the name of the Willow Garage nor the names of its
*     contributors may be used to endorse or promote products derived
*     from this software without specific prior written permission.
*
*  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
*  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
*  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
*  FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
*  COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
*  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
*  BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
*  LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
*  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
*  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
*  ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
*  POSSIBILITY OF SUCH DAMAGE.
*********************************************************************/

// Author(s): Matei Ciocarlie

#include "object_manipulator/grasp_execution/simple_grasp_executor.h"

//#include <demo_synchronizer/synchronizer_client.h>

namespace object_manipulator {

bool SimpleGraspExecutor::prepareGrasp(const object_manipulation_msgs::PickupGoal &pickup_goal,
				       const object_manipulation_msgs::Grasp &grasp)
{
  if (marker_publisher_) marker_publisher_->colorGraspMarker(marker_id_, 1.0, 0.0, 0.0); //red

  geometry_msgs::PoseStamped target_pose = getTargetPose(pickup_goal);
  geometry_msgs::PoseStamped grasp_pose = transformToModelFrame(target_pose, grasp.grasp_pose);
  grasp_pose.header.stamp = ros::Time(0);

  //demo_synchronizer::getClient().sync(3, "Checking IK solution for final grasp position and interpolated IK for lift");
  //demo_synchronizer::getClient().rviz(3, "Collision models;Interpolated IK;IK contacts;Grasp execution");

  if ( !mechInterface().getIKForPose(pickup_goal.arm_name, grasp_pose, ik_response_) ) 
  {
    ROS_INFO_STREAM("  Grasp execution: initial IK check failed");
    if (marker_publisher_) marker_publisher_->colorGraspMarker(marker_id_, 1.0, 1.0, 0.0); //yellow
    return false;
  }
  ROS_INFO("  Grasp executor: IK for grasp succeeded.");

  std::vector<double> grasp_joint_angles = ik_response_.solution.joint_state.position;
  if (!getInterpolatedIKForLift(pickup_goal, grasp, grasp_joint_angles, interpolated_lift_trajectory_))
  {
    ROS_INFO("  Grasp executor: failed to generate usable lift trajectory");
    if (marker_publisher_) marker_publisher_->colorGraspMarker(marker_id_, 0.0, 0.0, 1.0); //blue
    return false;
  }
  ROS_INFO("  Grasp executor: interpolated IK for lift succeeded.");
  
  return true;
}

GraspExecutor::Result 
SimpleGraspExecutor::executeGrasp(const object_manipulation_msgs::PickupGoal &pickup_goal,
				  const object_manipulation_msgs::Grasp &grasp)
{
  //demo_synchronizer::getClient().sync(2, "Using motion planner to move the arm to grasp position");
  //demo_synchronizer::getClient().rviz(1, "Collision models;Planning goal;Collision map;Environment contacts");

  mechInterface().handPostureGraspAction(pickup_goal.arm_name, grasp, 
					 object_manipulation_msgs::GraspHandPostureExecutionGoal::PRE_GRASP);

  if ( !mechInterface().attemptMoveArmToGoal(pickup_goal.arm_name, ik_response_.solution.joint_state.position) ) 
  {
    ROS_INFO("  Grasp execution: move arm reports failure");
    if (marker_publisher_) marker_publisher_->colorGraspMarker(marker_id_, 1.0, 0.5, 0.0); //orange-ish
    return UNFEASIBLE;
  }

  mechInterface().handPostureGraspAction(pickup_goal.arm_name, grasp, 
					 object_manipulation_msgs::GraspHandPostureExecutionGoal::GRASP);    

  if (marker_publisher_) marker_publisher_->colorGraspMarker(marker_id_, 0.0, 1.0, 0.0); //green
  return SUCCESS;
}

} //namespace object_manipulator
