/*********************************************************************
*
*  Copyright (c) 2009, Willow Garage, Inc.
*  All rights reserved.
*
*  Redistribution and use in source and binary forms, with or without
*  modification, are permitted provided that the following conditions
*  are met:
*
*   * Redistributions of source code must retain the above copyright
*     notice, this list of conditions and the following disclaimer.
*   * Redistributions in binary form must reproduce the above
*     copyright notice, this list of conditions and the following
*     disclaimer in the documentation and/or other materials provided
*     with the distribution.
*   * Neither the name of the Willow Garage nor the names of its
*     contributors may be used to endorse or promote products derived
*     from this software without specific prior written permission.
*
*  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
*  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
*  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
*  FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
*  COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
*  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
*  BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
*  LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
*  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
*  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
*  ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
*  POSSIBILITY OF SUCH DAMAGE.
*********************************************************************/

// Author(s): Matei Ciocarlie

#include "object_manipulator/object_manipulator.h"

//#include <demo_synchronizer/synchronizer_client.h>

#include "object_manipulator/grasp_execution/grasp_executor_with_approach.h"
#include "object_manipulator/grasp_execution/reactive_grasp_executor.h"
#include "object_manipulator/place_execution/place_executor.h"
#include "object_manipulator/tools/grasp_marker_publisher.h"
#include "object_manipulator/tools/exceptions.h"

using object_manipulation_msgs::GraspableObject;
using object_manipulation_msgs::PickupGoal;
using object_manipulation_msgs::PickupResult;
using object_manipulation_msgs::PlaceGoal;
using object_manipulation_msgs::PlaceResult;
using object_manipulation_msgs::ManipulationResult;

namespace object_manipulator {

ObjectManipulator::ObjectManipulator() :
  priv_nh_("~"),
  root_nh_(""),
  grasp_planning_services_("", "", false),
  marker_pub_(NULL)
{
  bool publish_markers = true;
  if (publish_markers)
  {
    marker_pub_ = new GraspMarkerPublisher();
  }

  grasp_executor_with_approach_ = new GraspExecutorWithApproach(marker_pub_);
  reactive_grasp_executor_ = new ReactiveGraspExecutor(marker_pub_);
  place_executor_ = new PlaceExecutor(marker_pub_);
  reactive_place_executor_ = new ReactivePlaceExecutor(marker_pub_);

  priv_nh_.param<std::string>("default_cluster_planner", default_cluster_planner_, "default_cluster_planner");
  priv_nh_.param<std::string>("default_database_planner", default_database_planner_, "default_database_planner");

  ROS_INFO("Object manipulator ready. Default cluster planner: %s. Default database planner: %s.", 
	   default_cluster_planner_.c_str(), default_database_planner_.c_str());
}

ObjectManipulator::~ObjectManipulator()
{
  delete marker_pub_;
  delete grasp_executor_with_approach_;
  delete reactive_grasp_executor_;
  delete place_executor_;
  delete reactive_place_executor_;
}

void ObjectManipulator::pickup(const PickupGoal::ConstPtr &pickup_goal,
			       actionlib::SimpleActionServer<object_manipulation_msgs::PickupAction> *action_server)
{
  //the result that will be returned
  PickupResult result;

  //populate a list of grasps to be tried
  std::vector<object_manipulation_msgs::Grasp> grasps;
  if (!pickup_goal->desired_grasps.empty())
  {
    //use the requested grasps, if any
    grasps = pickup_goal->desired_grasps;
  }
  else
  {
    //decide which grasp planner to call depending on the type of object
    std::string planner_service;
    switch (pickup_goal->target.type)
      {
      case GraspableObject::DATABASE_MODEL:
	planner_service = default_database_planner_;
	break;
      case GraspableObject::POINT_CLUSTER:
	planner_service = default_cluster_planner_;
	break;
      default:
	ROS_ERROR("Object manipulator: unknown type of graspable object");
	result.manipulation_result.value = ManipulationResult::ERROR;
	action_server->setAborted(result);
	return;
      }
    //call the planner and save the list of grasps
    object_manipulation_msgs::GraspPlanning srv;
    srv.request.arm_name = pickup_goal->arm_name;
    srv.request.target = pickup_goal->target;
    srv.request.collision_object_name = pickup_goal->collision_object_name;
    srv.request.collision_support_surface_name = pickup_goal->collision_support_surface_name;
    if (!grasp_planning_services_.client(planner_service).call(srv))
    {
      ROS_ERROR("Object manipulator failed to call planner at %s", planner_service.c_str());
      result.manipulation_result.value = ManipulationResult::ERROR;
      action_server->setAborted(result);
      return;
    }
    if (srv.response.error_code.value != srv.response.error_code.SUCCESS)
    {
      ROS_ERROR("Object manipulator: grasp planner failed with error code %d", srv.response.error_code.value);
      result.manipulation_result.value = ManipulationResult::ERROR;
      action_server->setAborted(result);
      return;
    }
    grasps = srv.response.grasps;
  }

  //decide which grasp executor will be used
  GraspExecutor *executor;
  if (pickup_goal->use_reactive_execution) executor = reactive_grasp_executor_;
  else executor = grasp_executor_with_approach_;

  //demo_synchronizer::getClient().sync(2, "Attempting to grasp an object.");
  //demo_synchronizer::getClient().rviz(1, "Collision models;Grasp execution;Interpolated IK");

  //try the grasps in the list until one succeeds
  try
  {
    for (size_t i=0; i<grasps.size(); i++)
    {
      if (action_server->isPreemptRequested())
      {
	action_server->setPreempted();
	return;
      }
      switch(executor->checkAndExecuteGrasp(*pickup_goal, grasps[i]))
      {
      case GraspExecutor::SUCCESS:	
	//demo_synchronizer::getClient().sync(2, "Object grasped - ready to move.");
	//demo_synchronizer::getClient().rviz(1, "Collision models;Planning goal;Collision map;Environment contacts");
	result.manipulation_result.value = ManipulationResult::SUCCESS;
	result.grasp = grasps[i];
	action_server->setSucceeded(result);
	return;
      case GraspExecutor::UNFEASIBLE:
	break;
      case GraspExecutor::FAILED:
	executor->retreat(*pickup_goal);
	result.manipulation_result.value = ManipulationResult::FAILED;
	result.grasp = grasps[i];
	action_server->setAborted(result);
	return;
      case GraspExecutor::LIFT_FAILED:
	result.manipulation_result.value = ManipulationResult::LIFT_FAILED;
	result.grasp = grasps[i];
	action_server->setAborted(result);
	return;
      default:
	ROS_ERROR("Object manipulator: unknown result code returned by executor");
	result.manipulation_result.value = ManipulationResult::ERROR;
	action_server->setAborted(result);
	return;
      }
    }
    //all the grasps have been deemed unfeasible
    result.manipulation_result.value = ManipulationResult::UNFEASIBLE;
    action_server->setAborted(result);
    return;
  }
  catch (MoveArmStuckException &ex)
  {
    ROS_INFO("Grasp aborted because move_arm is stuck");
    result.manipulation_result.value = ManipulationResult::ARM_MOVEMENT_PREVENTED;
    action_server->setAborted(result);
    return;
  }
  catch (GraspException &ex)
  {
    ROS_ERROR("Grasp error; exception: %s", ex.what());
    result.manipulation_result.value = ManipulationResult::ERROR;
    action_server->setAborted(result);
    return;
  }
}

void ObjectManipulator::place(const object_manipulation_msgs::PlaceGoal::ConstPtr &place_goal,
			      actionlib::SimpleActionServer<object_manipulation_msgs::PlaceAction> *action_server)
{
  PlaceResult result;
  PlaceExecutor *executor;
  if (place_goal->use_reactive_place)
  {
    executor = reactive_place_executor_;
  }
  else
  {
    executor = place_executor_;
  }
  try
  {
    switch (executor->place(*place_goal))
    {
    case PlaceExecutor::SUCCESS:
      result.manipulation_result.value = ManipulationResult::SUCCESS;
      action_server->setSucceeded(result);
      return;
    case PlaceExecutor::UNFEASIBLE:
      result.manipulation_result.value = ManipulationResult::UNFEASIBLE;
      action_server->setAborted(result);
      return;
    case PlaceExecutor::FAILED:
      result.manipulation_result.value = ManipulationResult::FAILED;
      action_server->setAborted(result);
      return;
    case PlaceExecutor::RETREAT_FAILED:
      result.manipulation_result.value = ManipulationResult::RETREAT_FAILED;
      action_server->setAborted(result);
      return;
    }
    result.manipulation_result.value = ManipulationResult::ERROR;
    action_server->setAborted(result);
    return;
  }
  catch (MoveArmStuckException &ex)
  {
    ROS_INFO("Place aborted because move_arm is stuck");
    result.manipulation_result.value = ManipulationResult::ARM_MOVEMENT_PREVENTED;
    action_server->setAborted(result);
    return;
  }
  catch (GraspException &ex)
  {
    ROS_ERROR("Place error; exception: %s", ex.what());
    result.manipulation_result.value = ManipulationResult::ERROR;
    action_server->setAborted(result);
    return;
  } 
}

}

