/*********************************************************************
*
*  Copyright (c) 2009, Willow Garage, Inc.
*  All rights reserved.
*
*  Redistribution and use in source and binary forms, with or without
*  modification, are permitted provided that the following conditions
*  are met:
*
*   * Redistributions of source code must retain the above copyright
*     notice, this list of conditions and the following disclaimer.
*   * Redistributions in binary form must reproduce the above
*     copyright notice, this list of conditions and the following
*     disclaimer in the documentation and/or other materials provided
*     with the distribution.
*   * Neither the name of the Willow Garage nor the names of its
*     contributors may be used to endorse or promote products derived
*     from this software without specific prior written permission.
*
*  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
*  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
*  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
*  FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
*  COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
*  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
*  BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
*  LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
*  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
*  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
*  ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
*  POSSIBILITY OF SUCH DAMAGE.
*********************************************************************/

#include "object_manipulator/tools/mechanism_interface.h"
#include "object_manipulator/tools/hand_description.h"
#include "object_manipulator/tools/exceptions.h"

namespace object_manipulator {

static const std::string IK_SERVICE_SUFFIX = "/constraint_aware_ik";
static const std::string INTERPOLATED_IK_SERVICE_SUFFIX = "/interpolated_ik";
static const std::string INTERPOLATED_IK_SET_PARAMS_SERVICE_SUFFIX = "/interpolated_ik_set_params";
static const std::string IK_QUERY_SERVICE_SUFFIX = "/get_ik_solver_info";
static const std::string CHECK_STATE_VALIDITY_SERVICE_SUFFIX = "/get_state_validity";
static const std::string GRASP_STATUS_SUFFIX = "/grasp_status";

static const std::string NORMALIZE_SERVICE_NAME = "trajectory_filter_unnormalizer/filter_trajectory";

static const std::string REACTIVE_GRASP_ACTION_SUFFIX = "/reactive_grasp";
static const std::string REACTIVE_LIFT_ACTION_SUFFIX = "/reactive_lift";
static const std::string REACTIVE_PLACE_ACTION_SUFFIX = "/reactive_place";
static const std::string MOVE_ARM_ACTION_SUFFIX = "/move_arm";
static const std::string TRAJECTORY_ACTION_SUFFIX = "/joint_trajectory";
static const std::string HAND_POSTURE_ACTION_SUFFIX = "/hand_posture_execution";

static const std::string MOVE_ARM_PLANNER_ID = "";
static const std::string MOVE_ARM_PLANNER_SERVICE_NAME = "ompl_planning/plan_kinematic_path";

static const std::string ATTACHED_COLLISION_TOPIC="attached_collision_object";

MechanismInterface::MechanismInterface() : 
  root_nh_(""),priv_nh_("~"),

  //------------------- multi arm service clients -----------------------
  ik_query_client_("", IK_QUERY_SERVICE_SUFFIX, true),
  ik_service_client_("", IK_SERVICE_SUFFIX, true),
  interpolated_ik_service_client_("", INTERPOLATED_IK_SERVICE_SUFFIX, true),
  interpolated_ik_set_params_client_("", INTERPOLATED_IK_SET_PARAMS_SERVICE_SUFFIX, true),
  check_state_validity_client_("", CHECK_STATE_VALIDITY_SERVICE_SUFFIX, true),
  grasp_status_client_("", GRASP_STATUS_SUFFIX, true),
  //------------------- simple service clients -----------------------  
  joint_trajectory_normalizer_service_(NORMALIZE_SERVICE_NAME),
  //-------------------- multi arm action clients -----------------------
  reactive_grasp_action_client_("", REACTIVE_GRASP_ACTION_SUFFIX, true, true),
  reactive_lift_action_client_("", REACTIVE_LIFT_ACTION_SUFFIX, true, true),
  reactive_place_action_client_("", REACTIVE_PLACE_ACTION_SUFFIX, true, true),
  move_arm_action_client_("", MOVE_ARM_ACTION_SUFFIX, true, true),
  traj_action_client_("", TRAJECTORY_ACTION_SUFFIX, true, true),
  hand_posture_client_("", HAND_POSTURE_ACTION_SUFFIX, true, true)
{
  //collision map publishing topic
  attached_object_pub_ = root_nh_.advertise<mapping_msgs::AttachedCollisionObject>(ATTACHED_COLLISION_TOPIC, 10);
}

/*! For now, just calls the IK Info service each time. In the future, we might do some
 caching in here.
*/
std::vector<std::string> MechanismInterface::getJointNames(std::string arm_name)
{
  kinematics_msgs::GetKinematicSolverInfo::Request query_request;
  kinematics_msgs::GetKinematicSolverInfo::Response query_response;  
  if ( !ik_query_client_.client(arm_name).call(query_request, query_response) ) 
  {
    ROS_ERROR("Failed to call ik information query");
    throw MechanismException("Failed to call ik information query");
  }
  return query_response.kinematic_solver_info.joint_names;
}

void MechanismInterface::attemptTrajectory(std::string arm_name, 
					   const trajectory_msgs::JointTrajectory &trajectory, 
					   bool unnormalize)
{
  if (trajectory.points.empty()) 
  {
    ROS_ERROR("attemptTrajectory called with empty trajectory");
    throw MechanismException("attemptTrajectory called with empty trajectory");
  }

  pr2_controllers_msgs::JointTrajectoryGoal goal;
  if (unnormalize)
  {
    motion_planning_msgs::FilterJointTrajectory service_call;
    service_call.request.trajectory = trajectory;
    service_call.request.allowed_time = ros::Duration(2.0);
    if ( !joint_trajectory_normalizer_service_.client().call(service_call) )
    {
      ROS_ERROR("Mechanism interface: joint trajectory normalizer service call failed");
      throw MechanismException("joint trajectory normalizer service call failed");
    } 
    goal.trajectory = service_call.response.trajectory;
  }
  else
  {
    goal.trajectory = trajectory;
  }
  goal.trajectory.header.stamp = ros::Time::now() + ros::Duration(1.0);

  //wait 5 seconds more that the whole trajectory is supposed to take
  ros::Duration timeout = ros::Duration(1.0) + trajectory.points.back().time_from_start + ros::Duration(5.0);
  traj_action_client_.client(arm_name).sendGoal(goal);
  if ( !traj_action_client_.client(arm_name).waitForResult(timeout) ) 
  {
    ROS_ERROR("  Trajectory timed out");
    throw MechanismException("trajectory timed out");
  }
}

void MechanismInterface::setInterpolatedIKParams(std::string arm_name, int num_steps,
						 int collision_check_resolution, bool start_from_end)
{
  interpolated_ik_motion_planner::SetInterpolatedIKMotionPlanParams srv;
  srv.request.num_steps = num_steps;
  srv.request.consistent_angle = M_PI/9;
  srv.request.collision_check_resolution = collision_check_resolution;
  srv.request.steps_before_abort = 0;
  srv.request.pos_spacing = 0.01; //ignored if num_steps !=0
  srv.request.rot_spacing = 0.1;  //ignored if num_steps !=0
  srv.request.collision_aware = true;
  srv.request.start_from_end = start_from_end;
  if (!interpolated_ik_set_params_client_.client(arm_name).call(srv))
  {
    ROS_ERROR("Failed to set Interpolated IK server parameters");
    throw MechanismException("Failed to set Interpolated IK server parameters");
  }
}

bool MechanismInterface::moveArmToPose(std::string arm_name, const geometry_msgs::PoseStamped &desired_pose)
{
  kinematics_msgs::GetConstraintAwarePositionIK::Response ik_response;
  if(!getIKForPose(arm_name, desired_pose,ik_response))
    return false;
  if(!attemptMoveArmToGoal(arm_name, ik_response.solution.joint_state.position))
    return false;
  return true;
}

bool MechanismInterface::getIKForPose(std::string arm_name, const geometry_msgs::PoseStamped &desired_pose,
				      kinematics_msgs::GetConstraintAwarePositionIK::Response& ik_response)
{
  //call collision-aware ik
  kinematics_msgs::GetConstraintAwarePositionIK::Request ik_request;
  ik_request.ik_request.ik_link_name = handDescription().gripperFrame(arm_name);
  ik_request.ik_request.pose_stamped.pose = desired_pose.pose;
  ik_request.ik_request.pose_stamped.header.stamp = desired_pose.header.stamp;
  ik_request.ik_request.pose_stamped.header.frame_id = desired_pose.header.frame_id;
  ik_request.ik_request.ik_seed_state.joint_state.name = getJointNames(arm_name);
  ik_request.ik_request.ik_seed_state.joint_state.position.resize(7, 0.0);
  ik_request.timeout = ros::Duration(2.0);
  if( !ik_service_client_.client(arm_name).call(ik_request,ik_response) ) 
  {
    ROS_ERROR("IK Service Call failed altogether");
    throw MechanismException("IK Service Call failed altogether");
  }
  if (ik_response.error_code.val != ik_response.error_code.SUCCESS)
  {
    ROS_INFO("Get IK failed with error code %d", ik_response.error_code.val);
    return false;
  }
  return true;
}

bool MechanismInterface::checkStateValidity(std::string arm_name, const std::vector<double> &joint_values)
{
  planning_environment_msgs::GetStateValidity::Request req;
  planning_environment_msgs::GetStateValidity::Response res;

  req.robot_state.joint_state.name = getJointNames(arm_name);
  req.robot_state.joint_state.position = joint_values;
  if ( req.robot_state.joint_state.name.size() != joint_values.size() )
  {
    throw MechanismException("Wrong number of joint values for checkStateValidity");
  }
  req.robot_state.joint_state.header.stamp = ros::Time::now();
  req.check_collisions = true;

  if(!check_state_validity_client_.client(arm_name).call(req,res))
  {
    throw MechanismException("Call to check state validity client failed");
  }

  if (res.error_code.val == res.error_code.SUCCESS) return true;
  return false;
}

/*! If starting_from_end is set to true, the Interpolated IK server will be set to start 
  from end as well, and the resulting trajectory is copied into the result starting from 
  the end until a failed step is encountered.

  - seed_joint_position is a seed to be use for IK for the joints we are planning on. Pass
  an empty vector if you don't have a seed you want to use

  - joint_state is a list of values to be used for joints that are not part of our plan.
  For example, use this to specifiy if you want the plan done with the gripper open
  or closed. If you don't specify a joint in here, the current value of that joint will
  be used by the Interpolated IK server.
 */
void MechanismInterface::getInterpolatedIK(std::string arm_name,
					   geometry_msgs::PoseStamped start_pose,
					   geometry_msgs::Vector3Stamped direction,
					   float desired_trajectory_length,
					   const std::vector<double> &seed_joint_position,
					   const sensor_msgs::JointState &joint_state,
					   const motion_planning_msgs::OrderedCollisionOperations &collision_operations,
					   const std::vector<motion_planning_msgs::LinkPadding> &link_padding,
					   bool reverse_trajectory,
					   trajectory_msgs::JointTrajectory &trajectory,
					   float &actual_trajectory_length)
{
  //first compute the desired end pose
  //make sure the input is normalized
  geometry_msgs::Vector3Stamped direction_norm = direction;
  direction_norm.vector = normalize(direction.vector);
  //multiply by the length
  desired_trajectory_length = fabs(desired_trajectory_length);
  direction_norm.vector.x *= desired_trajectory_length;
  direction_norm.vector.y *= desired_trajectory_length;
  direction_norm.vector.z *= desired_trajectory_length;
  geometry_msgs::PoseStamped end_pose = translateGripperPose(direction_norm, start_pose, arm_name);
 
  //hard-coded for now
  float max_step_size = 0.01;
  unsigned int collision_check_resolution = 2;
 
  //compute the number of steps  
  unsigned int num_steps = (unsigned int)ceil(desired_trajectory_length / fabs(max_step_size));
  float actual_step_size = desired_trajectory_length / num_steps;

  ROS_INFO("Trajectory details: length %f, requested num steps: %d, actual step size: %f",
	   desired_trajectory_length, num_steps, actual_step_size);

  if (reverse_trajectory)
  {
    std::swap(start_pose, end_pose);
  }

  //recall that here we setting the numbre of points in trajectory, which is steps+1
  setInterpolatedIKParams(arm_name, num_steps+1, collision_check_resolution, reverse_trajectory);

  motion_planning_msgs::RobotState start_state;
  start_state.multi_dof_joint_state.child_frame_id = handDescription().gripperFrame(arm_name);
  start_state.multi_dof_joint_state.pose = start_pose.pose;
  start_state.multi_dof_joint_state.frame_id = start_pose.header.frame_id;
  start_state.multi_dof_joint_state.stamp = ros::Time::now();

  //pass the seeds for the IK, if any
  if (!seed_joint_position.empty())
  { 
    //we are silently assuming that the values passed in match out joint names for IK
    start_state.joint_state.name = getJointNames(arm_name);
    if (seed_joint_position.size() != start_state.joint_state.name.size())
    {
      ROS_ERROR("Interpolated IK request: seed_joint_position does not match joint names");
      throw MechanismException("Interpolated IK request: seed_joint_position does not match joint names");
    }
    start_state.joint_state.position = seed_joint_position;
  }

  //pass the desired values of non-planned joints, if any
  for (size_t i=0; i<joint_state.name.size(); i++)
  {
    start_state.joint_state.name.push_back(joint_state.name[i]);
    start_state.joint_state.position.push_back(joint_state.position[i]);
  }
  
  motion_planning_msgs::PositionConstraint position_constraint;
  motion_planning_msgs::OrientationConstraint orientation_constraint;
  motion_planning_msgs::poseStampedToPositionOrientationConstraints(end_pose, handDescription().gripperFrame(arm_name),
								    position_constraint, 
								    orientation_constraint);
  motion_planning_msgs::Constraints goal_constraints;
  goal_constraints.position_constraints.push_back(position_constraint);
  goal_constraints.orientation_constraints.push_back(orientation_constraint);

  motion_planning_msgs::GetMotionPlan motion_plan;
  motion_plan.request.motion_plan_request.start_state = start_state;
  motion_plan.request.motion_plan_request.goal_constraints = goal_constraints;
  motion_plan.request.motion_plan_request.ordered_collision_operations = collision_operations;

  //pass the dynamic link padding, if any
  motion_plan.request.motion_plan_request.link_padding = link_padding;

  if ( !interpolated_ik_service_client_.client(arm_name).call(motion_plan) ) 
  {
    ROS_ERROR("  Call to Interpolated IK service failed");
    throw MechanismException("Call to Interpolated IK service failed");
  }

  trajectory.points.clear();
  trajectory.joint_names = motion_plan.response.trajectory.joint_trajectory.joint_names;

  if (motion_plan.response.trajectory_error_codes.empty()) 
  {
    ROS_ERROR("  Interpolated IK: empty trajectory received");
    throw MechanismException("Interpolated IK: empty trajectory received");
  }
  
  if (!reverse_trajectory)
  {
    for (size_t i=0; i<motion_plan.response.trajectory_error_codes.size(); i++) 
    {
      if ( motion_plan.response.trajectory_error_codes[i].val == motion_plan.response.trajectory_error_codes[i].SUCCESS )
      {
	trajectory.points.push_back( motion_plan.response.trajectory.joint_trajectory.points[i] );
      } 
      else 
      {
	ROS_INFO("  Interpolated IK failed on step %d (forward towards %d) with error code %d", 
		 (int) i, 
		 (int) motion_plan.response.trajectory_error_codes.size() - 1, 
		 motion_plan.response.trajectory_error_codes[i].val);
	break;
      }
    }
  }
  else
  {
    size_t first_success = 0;
    while ( first_success < motion_plan.response.trajectory_error_codes.size() &&
	    motion_plan.response.trajectory_error_codes[first_success].val != 
	    motion_plan.response.trajectory_error_codes[first_success].SUCCESS ) first_success ++;
    if (first_success != 0)
    {
      ROS_INFO("  Interpolation failed on step %d (backwards from %d) with error code %d",
	       (int) first_success - 1,
	       (int) motion_plan.response.trajectory_error_codes.size() - 1,
	       motion_plan.response.trajectory_error_codes[first_success - 1].val);
    }
    else
    {
      ROS_INFO("  Interpolation trajectory complete (backwards from %d points)",
	       (int) motion_plan.response.trajectory_error_codes.size());
      
    }
    for (size_t i=first_success; i < motion_plan.response.trajectory_error_codes.size(); i++) 
    {
      if ( motion_plan.response.trajectory_error_codes[i].val == motion_plan.response.trajectory_error_codes[i].SUCCESS )
      {
	trajectory.points.push_back( motion_plan.response.trajectory.joint_trajectory.points[i] );
      } 
      else 
      {
	ROS_ERROR("  Interpolated IK: unexpected behavior for error codes: step %d has error code %d",
		  (int) i, motion_plan.response.trajectory_error_codes[i].val);
	throw MechanismException("Interpolated IK: unexpected behavior for error codes");
      }
    }
  }

  if (!trajectory.points.empty()) actual_trajectory_length = actual_step_size * (trajectory.points.size()-1);
  else actual_trajectory_length = 0.0;
}

bool MechanismInterface::attemptMoveArmToGoal(std::string arm_name, 
                                              const std::vector<double> &desired_joint_values) 
{
  int num_tries = 0;
  int max_tries = 5;
  move_arm_msgs::MoveArmGoal move_arm_goal;

  move_arm_goal.motion_plan_request.group_name = handDescription().armGroup(arm_name);
  move_arm_goal.motion_plan_request.num_planning_attempts = 1;
  move_arm_goal.motion_plan_request.allowed_planning_time = ros::Duration(5.0);
  move_arm_goal.motion_plan_request.planner_id = MOVE_ARM_PLANNER_ID;
  move_arm_goal.planner_service_name = MOVE_ARM_PLANNER_SERVICE_NAME;
  //move_arm_goal.disable_collision_monitoring = true;

  std::vector<std::string> joint_names = getJointNames(arm_name);
  move_arm_goal.motion_plan_request.goal_constraints.joint_constraints.resize(joint_names.size());
  for(unsigned int i = 0; i < desired_joint_values.size(); i++) 
  {
    move_arm_goal.motion_plan_request.goal_constraints.joint_constraints[i].joint_name = joint_names.at(i);
    move_arm_goal.motion_plan_request.goal_constraints.joint_constraints[i].position = desired_joint_values[i];
    move_arm_goal.motion_plan_request.goal_constraints.joint_constraints[i].tolerance_below = .08;
    move_arm_goal.motion_plan_request.goal_constraints.joint_constraints[i].tolerance_above = .08;
  }
  
  bool success = false;
  motion_planning_msgs::ArmNavigationErrorCodes error_code;
  while(num_tries < max_tries)
  {
    move_arm_action_client_.client(arm_name).sendGoal(move_arm_goal);
    bool withinWait = move_arm_action_client_.client(arm_name).waitForResult(ros::Duration(60.0));
    if(!withinWait) 
    {
      move_arm_action_client_.client(arm_name).cancelGoal();
      ROS_INFO("   Move arm goal could not be achieved by move_arm in the allowed duration");
      success = false;
      num_tries++;
      move_arm_msgs::MoveArmResult move_arm_result = *move_arm_action_client_.client(arm_name).getResult();
      error_code.val = error_code.TIMED_OUT;
      modifyMoveArmGoal(move_arm_goal,error_code,move_arm_result.contacts);
      continue;
    }
    actionlib::SimpleClientGoalState state = move_arm_action_client_.client(arm_name).getState();
    if(state == actionlib::SimpleClientGoalState::SUCCEEDED) 
    {
      ROS_INFO("  Move arm: position was successfully achieved");
      success = true;
      break;
    } 
    else 
    {
      move_arm_msgs::MoveArmResult move_arm_result = *move_arm_action_client_.client(arm_name).getResult();
      ROS_INFO("   Move arm: non-success state was reached with id: %d",move_arm_result.error_code.val);
      ROS_INFO("   num_tries: %d, max_tries: %d",num_tries,max_tries);
      success = false;
      num_tries++;
      error_code = move_arm_result.error_code;
      modifyMoveArmGoal(move_arm_goal,error_code,move_arm_result.contacts);
      continue;
    }
  }
  //! If move arm is stuck, then nothing will get it out of here
  if (!success && error_code.val == error_code.START_STATE_IN_COLLISION)
  {
    throw MoveArmStuckException();
  }
  return success;
}


void MechanismInterface::modifyMoveArmGoal(move_arm_msgs::MoveArmGoal &move_arm_goal, 
                                           motion_planning_msgs::ArmNavigationErrorCodes &error_code,
                                           std::vector<planning_environment_msgs::ContactInformation> &contact_info_)
{
  double allowed_penetration_depth = 0.03;
  if(error_code.val == error_code.START_STATE_IN_COLLISION)
  {
    std::vector<motion_planning_msgs::AllowedContactSpecification> allowed_contacts;
    for(unsigned int i=0; i < contact_info_.size(); i++)
    {
      if(contact_info_[i].depth < allowed_penetration_depth)
      {
        motion_planning_msgs::AllowedContactSpecification allowed_contact_tmp;
        allowed_contact_tmp.shape.type = allowed_contact_tmp.shape.BOX;
        allowed_contact_tmp.shape.dimensions.resize(3);
        allowed_contact_tmp.shape.dimensions[0] = 0.03;
        allowed_contact_tmp.shape.dimensions[1] = 0.03;
        allowed_contact_tmp.shape.dimensions[2] = 0.03;
        allowed_contact_tmp.pose_stamped.pose.position = contact_info_[i].position;
        allowed_contact_tmp.pose_stamped.pose.orientation.x = 0.0;
        allowed_contact_tmp.pose_stamped.pose.orientation.y = 0.0;
        allowed_contact_tmp.pose_stamped.pose.orientation.z = 0.0;
        allowed_contact_tmp.pose_stamped.pose.orientation.w = 1.0;
        allowed_contact_tmp.pose_stamped.header.stamp = ros::Time::now();
        allowed_contact_tmp.pose_stamped.header.frame_id = contact_info_[i].header.frame_id;
        allowed_contact_tmp.link_names.push_back(contact_info_[i].contact_body_1);
        allowed_contact_tmp.penetration_depth = allowed_penetration_depth;
        allowed_contacts.push_back(allowed_contact_tmp);
        ROS_INFO("Added allowed contact region: %d",i);
        ROS_INFO("Position                    : (%f,%f,%f)",contact_info_[i].position.x,
		 contact_info_[i].position.y,contact_info_[i].position.z);
        ROS_INFO("Frame id                    : %s",contact_info_[i].header.frame_id.c_str());
        ROS_INFO("Depth                       : %f",allowed_penetration_depth);
        ROS_INFO("Link                        : %s",contact_info_[i].contact_body_1.c_str());
        ROS_INFO(" ");
      }
    }
    ROS_INFO("Added %d allowed contact regions",(int)allowed_contacts.size());
    move_arm_goal.motion_plan_request.allowed_contacts = allowed_contacts;
  }
}

/*! The translation must be expressed either in the gripper frame or in the robot frame. The reason is the 
  following: this function must work even without tf, so that it works for hypothetical situations,
  or situations in which the complete tf tree is unknown. Its only decision is thus whether to
  pre-multiply or post-multiply the pose with the given translation. 

  If the translation is in gripper frame, the pose gets post-multiplied by it.
  If the translation is in robot frame, the pose gets pre-multiplied by it.

  The input pose is first transformed to the robor frame, then translated, then transformed back into
  the original frame.
 */
geometry_msgs::PoseStamped MechanismInterface::translateGripperPose(geometry_msgs::Vector3Stamped translation, 
								    geometry_msgs::PoseStamped start_pose,
								    std::string arm_name)
{
  bool pre_multiply;
  if (translation.header.frame_id == handDescription().gripperFrame(arm_name))
  {
    pre_multiply=false;
  } 
  else if (translation.header.frame_id == handDescription().robotFrame(arm_name))
  {
    pre_multiply=true;
  }
  else
  {
    throw MechanismException(std::string("Gripper translations (such as for lift or approach) can only be specified in "
					 "either the gripper frame or the robot frame. This one was specified in frame ") + 
			     translation.header.frame_id);
  }


  //go to robot frame first
  geometry_msgs::PoseStamped start_pose_robot_frame = transformPose(handDescription().robotFrame(arm_name), start_pose);
  tf::StampedTransform start_transform;
  tf::poseMsgToTF(start_pose_robot_frame.pose, start_transform);

  btVector3 vec;
  tf::vector3MsgToTF(translation.vector, vec);
  tf::StampedTransform translate_trans;
  translate_trans.setIdentity();
  translate_trans.setOrigin(vec);

  //compute the translated pose
  tf::Transform end_transform;
  if (pre_multiply) end_transform = translate_trans * start_transform;
  else end_transform = start_transform * translate_trans;

  //prepare the results
  geometry_msgs::PoseStamped translated_pose;
  tf::poseTFToMsg(end_transform, translated_pose.pose);
  translated_pose.header.frame_id = start_pose.header.frame_id;
  translated_pose.header.stamp = ros::Time::now();

  //return the result in the requested frame
  translated_pose = transformPose(start_pose.header.frame_id, translated_pose);

  return translated_pose;
}

/*! Current gripper pose is returned in the requested frame_id.*/
geometry_msgs::PoseStamped MechanismInterface::getGripperPose(std::string arm_name, std::string frame_id)
{
  tf::StampedTransform gripper_transform;
  try
  {
    listener_.lookupTransform(frame_id, handDescription().gripperFrame(arm_name), ros::Time(0), gripper_transform);
  }
  catch (tf::TransformException ex)
  {
    ROS_ERROR("Mechanism interface: failed to get tf transform for wrist roll link; tf exception %s", ex.what());
    throw MechanismException(std::string("failed to get tf transform for wrist roll link; tf exception: ") + 
			     std::string(ex.what()) );
  }
  geometry_msgs::PoseStamped gripper_pose;
  tf::poseTFToMsg(gripper_transform, gripper_pose.pose);
  gripper_pose.header.frame_id = frame_id;
  gripper_pose.header.stamp = ros::Time::now();
  return gripper_pose;
}

void MechanismInterface::transformPointCloud(std::string target_frame, 
					     const sensor_msgs::PointCloud &cloud_in,
					     sensor_msgs::PointCloud &cloud_out)
{
  try
  {
    listener_.transformPointCloud(target_frame, cloud_in, cloud_out);    
  }
  catch (tf::TransformException ex)
  {
    ROS_ERROR("Mechanism interface: failed to cloud into %s frame; exception: %s", target_frame.c_str(),
	      ex.what());
    throw MechanismException(std::string("failed to transform cloud into frame ") + target_frame + 
			     std::string("; tf exception: ") + std::string(ex.what()) );
  }
}

geometry_msgs::PoseStamped MechanismInterface::transformPose(const std::string target_frame, 
							     const geometry_msgs::PoseStamped &stamped_in)
{
  geometry_msgs::PoseStamped stamped_out;
  try
  {
    listener_.transformPose(target_frame, stamped_in, stamped_out);
  }
  catch (tf::TransformException ex)
  {
    ROS_ERROR("Mechanism interface: failed to transform pose into %s frame; exception: %s", target_frame.c_str(),
	      ex.what());
    throw MechanismException(std::string("failed to transform pose into frame ") + target_frame + 
			     std::string("; tf exception: ") + std::string(ex.what()) );
  }
  return stamped_out;
}

/*! Moves the gripper from its current pose to the one obtained by the specified translation.
*/
bool MechanismInterface::translateGripper(std::string arm_name, const geometry_msgs::Vector3Stamped &direction,
					  motion_planning_msgs::OrderedCollisionOperations ord, 
					  const std::vector<motion_planning_msgs::LinkPadding> &link_padding,
					  float requested_distance, float min_distance,
					  float &actual_distance)
{
  //get the current gripper pose in the robot frame
  geometry_msgs::PoseStamped start_pose_stamped = getGripperPose(arm_name, handDescription().robotFrame(arm_name));

  //compute the interpolated trajectory
  trajectory_msgs::JointTrajectory traj;
  std::vector<double> empty;
  sensor_msgs::JointState empty_state;
  getInterpolatedIK(arm_name, 
		    start_pose_stamped, direction, requested_distance, 
		    empty, empty_state, ord, link_padding, false, traj, actual_distance);

  if (min_distance > 0 && actual_distance < min_distance) 
  {
    ROS_INFO("Mechanism interface: interpolated IK trajectory covers %f distance, but at least %f requested", 
	     actual_distance, min_distance);
    actual_distance = 0;
    return false;
  }

  if (traj.points.empty())
  {
    ROS_INFO("Mechanism interface: translate gripper trajectory is empty");
    actual_distance = false;
    return false;
  }

  //execute the normalized interpolated trajectory
  attemptTrajectory(arm_name, traj, true);
  return true;
}

/*! Object pose is obtained by multiplying the grasp_pose with the current pose
  of the wrist roll link of the given arm. Result is given in base_link frame.
*/
geometry_msgs::PoseStamped MechanismInterface::getObjectPoseForGrasp(std::string arm_name, 
								     const geometry_msgs::Pose &grasp_pose)

{
  //get the current pose of the gripper in base link coordinate frame
  tf::StampedTransform wrist_transform;
  try
  {
    listener_.lookupTransform("base_link",handDescription().gripperFrame(arm_name), 
			      ros::Time(0), wrist_transform);
  }
  catch (tf::TransformException ex)
  {
    ROS_ERROR("Mechanism interface: failed to get tf transform for wrist roll link");
    throw MechanismException(std::string("failed to get tf transform for wrist roll link; tf exception: ") + 
			     std::string(ex.what()) );
  }

  //multiply by inverse of grasp pose
  tf::Transform grasp_transform;
  tf::poseMsgToTF(grasp_pose, grasp_transform);

  tf::Transform object_transform;
  object_transform = wrist_transform * grasp_transform.inverse();

  //prepare the result
  geometry_msgs::PoseStamped object_pose;
  tf::poseTFToMsg(object_transform, object_pose.pose);
  object_pose.header.frame_id = "base_link";
  object_pose.header.stamp = ros::Time::now();
  return object_pose;
}

void MechanismInterface::attachObjectToGripper(std::string arm_name, std::string collision_object_name)
{
  mapping_msgs::AttachedCollisionObject obj;
  obj.object.header.stamp = ros::Time::now();
  obj.object.header.frame_id = handDescription().robotFrame(arm_name);
  obj.object.operation.operation = mapping_msgs::CollisionObjectOperation::ATTACH_AND_REMOVE_AS_OBJECT;
  obj.object.id = collision_object_name;
  obj.link_name = handDescription().attachLinkName(arm_name);
  obj.touch_links = handDescription().gripperTouchLinkNames(arm_name);
  attached_object_pub_.publish(obj);
}
 
void MechanismInterface::detachAndAddBackObjectsAttachedToGripper(std::string arm_name, 
								  std::string collision_object_name)
{
  mapping_msgs::AttachedCollisionObject att;
  att.object.header.stamp = ros::Time::now();
  att.object.header.frame_id = handDescription().robotFrame(arm_name);
  att.link_name = handDescription().attachLinkName(arm_name);
  att.object.id = collision_object_name;
  att.object.operation.operation = mapping_msgs::CollisionObjectOperation::DETACH_AND_ADD_AS_OBJECT;
  attached_object_pub_.publish(att);
}

void MechanismInterface::handPostureGraspAction(std::string arm_name, 
						const object_manipulation_msgs::Grasp &grasp, int goal)
{
  object_manipulation_msgs::GraspHandPostureExecutionGoal posture_goal;
  posture_goal.grasp = grasp;
  posture_goal.goal = goal;
  hand_posture_client_.client(arm_name).sendGoal(posture_goal);
  bool withinWait = hand_posture_client_.client(arm_name).waitForResult(ros::Duration(10.0));
  if(!withinWait) 
  {
    hand_posture_client_.client(arm_name).cancelGoal();
    ROS_ERROR("Hand posture controller timed out on goal (%d)", goal);
    throw MechanismException("Hand posture controller timed out");
  }
  actionlib::SimpleClientGoalState state = hand_posture_client_.client(arm_name).getState();
  if(state != actionlib::SimpleClientGoalState::SUCCEEDED) 
  {
    ROS_ERROR("Hand posture controller failed on goal (%d)", goal);
    throw MechanismException("Hand posture controller failed");
  }
  ROS_INFO("Hand posture controller successfully achieved goal %d", goal);
}

bool MechanismInterface::graspPostureQuery(std::string arm_name)
{
  object_manipulation_msgs::GraspStatus query;
  if (!grasp_status_client_.client(arm_name).call(query))
  {
    ROS_ERROR("Grasp posture query call failed");
    throw MechanismException("Grasp posture query call failed");
  }
  return query.response.is_hand_occupied;
}


} //namespace object_manipulator
